import base64url = require("base64url")
import debug = require("debug")
import * as jwt from "jsonwebtoken"

import * as util from "util"
import {shim} from "util.promisify"

import { MissingKeyIdError } from "./MissingKeyIdError"
import { UnknownKeyIdError } from "./UnknownKeyIdError"

export type PubkeyData = { cert: string, alg: string } | undefined | null
export type PrivkeyData = { key: string, passphrase: string, alg: string } | undefined | null

export type PubkeyResolver = (keyId: string) => PubkeyData | Promise<PubkeyData>
export type PrivkeyResolver = (keyId: string) => PrivkeyData | Promise<PrivkeyData>

type JwtVerifyAsync = (token: string, publicKey: string, options?: jwt.VerifyOptions) => Promise<object | string>
type JwtSignAsync = (payload: object, privateKey: {}, options?: jwt.SignOptions) => Promise<string>

shim() // util.promisify shim

export class JwtHandler {

  private debug: debug.IDebugger

  private pubkeyResolver: PubkeyResolver | null
  private privkeyResolver: PrivkeyResolver | null

  private jwtVerifyAsync = util.promisify(jwt.verify) as JwtVerifyAsync
  private jwtSignAsync = util.promisify(jwt.sign) as JwtSignAsync

  public constructor(debugNamePrefix: string,
                     pubkeyResolver: PubkeyResolver | null,
                     privkeyResolver: PrivkeyResolver | null) {
    this.debug = debug(debugNamePrefix + ":jwt.handler")
    this.pubkeyResolver = pubkeyResolver
    this.privkeyResolver = privkeyResolver
  }

  /**
   * Extract key ID from the given JWT
   *
   * @param  {type} jwtRaw The JWT in raw form, i.e. Base64 coded parts separated with dots
   * @return {Promise<string, MissingKeyIdError>} Promise to the key id
   */
  public extractKeyId(jwtRaw: string): string {
    try {
      const jwtHeaderBase64 = jwtRaw.split(".", 1)
      const jwtHeader = JSON.parse(base64url.decode(jwtHeaderBase64[0]))
      if (jwtHeader.kid) {
        return jwtHeader.kid
      }
      throw new MissingKeyIdError()
    } catch (err) {
      if (!(err instanceof MissingKeyIdError)) {
        throw new jwt.JsonWebTokenError("JWT header parsing error", err)
      }
      throw err
    }
  }

  /**
   * Validates the given JWT
   *
   * @param {string} jwtRaw The JWT in raw form, i.e. Base64 coded parts separated with dots
   * @param {Object} options Validation options (jsonwebtoken module options)
   * @return {Promise<Object, JsonWebTokenError>} Promise to the JWT body
   */
  public async verify(jwtRaw: string, options?: jwt.VerifyOptions): Promise<string | object> {
    if (!jwtRaw) {
      throw new jwt.JsonWebTokenError("Empty JWT")
    }
    const keyId = await this.extractKeyId(jwtRaw)
    debug("verify, key id: " + keyId)
    if (this.pubkeyResolver) {
      const certData = await this.pubkeyResolver(keyId)
      if (!certData) {
        throw new UnknownKeyIdError(keyId)
      }
      debug("cert found")
      return this.jwtVerifyAsync(jwtRaw, certData.cert, options)
    }
    throw new Error("No public key resolver specified")
  }

  /**
   * Creates a new JWT with the given body and signs it with the given key
   *
   * @param {string} tokenBody The body of the JWT token
   * @param {string} keyId The ID of the signing key
   * @return {Promise<Object, JsonWebTokenError>} Promise to the JWT body
   */
  public async create(tokenBody: object, keyId: string): Promise<string> {
    debug("create, key id: " + keyId)
    if (this.privkeyResolver) {
      const signingKey = await this.privkeyResolver(keyId)
      if (!signingKey) {
        throw new UnknownKeyIdError("Unknown key id")
      }
      debug("priv key found")
      return this.jwtSignAsync(tokenBody, signingKey, { algorithm: signingKey.alg, header: { kid: keyId } })
    }
    throw new Error("No private key resolver specified")
  }

}
